-- A state machine
import "CoreLibs/object"

class("StateMachine").extends()

function StateMachine.new(params, stateTable)
   local fsm = StateMachine(params, stateTable)
   return fsm
end

-- params is a table of parameters to always pass to the functions
-- stateTable is a table of initial states. Entries should be of the form:
--   name = {main=mainFunc, incoming=inFunc, outgoing=outFunc}
--   where mainFunc and transFunc are functions that should be executed in relation to the state.
--   the State Machine will call mainFunc on execute()
--   the State Machine will call inFunc when the state is entered
--   the State Machine will call outFunc when the state in exited
function StateMachine:init(params, stateTable)
   self.currentState = nil
   self.parameters = params or {}
   self.states = stateTable or {}
end

function StateMachine:addState(name, mainFunc, inFunc, outFunc)
   self.states[name] = {main=mainFunc, incoming=inFunc, outgoing=outFunc}
end

-- execute the main function for the current state.
-- returns true if the function was executed, false otherwise
function StateMachine:execute()
   if self.currentState == nil or not self.states[self.currentState]
      or not self.states[self.currentState].main then
      return false
   end

   self.states[self.currentState].main(table.unpack(self.parameters))
   return true
end

-- change to the specified state. returns true if the state change is successful
function StateMachine:changeState(state)
   if not self.states[state] then
      print("WARN (StateMachine): tried to enter nonexistent state " .. state)
      return false
   end

   -- run the exit function for the previous state
   if self.currentState and self.states[self.currentState].outgoing then
      self.states[self.currentState].outgoing(table.unpack(self.parameters))
   end

   -- change state
   self.currentState = state

   -- run the entrance function for the new state
   if self.states[state].incoming then
      self.states[state].incoming(table.unpack(self.parameters))
   end
   return true
end
